/* 
 * Copyright 2015-2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

/****************************************************************************
   Program:     $Id: messages.cc,v 1.32 2016/06/29 22:55:18 kkeys Exp $
   Author:      Ken Keys, CAIDA
   Date:        $Date: 2016/06/29 22:55:18 $
   Description: Spoofer control channel message routines
****************************************************************************/
#include "spoof.h"

/**
 * Like inet_ntop() for an address stored in a std::string (as in a spoofer
 * protobuf).
 *
 * If the address size is incorrect for the family, the address will be dumped
 * in hex, making this function safe to use even when the source of the
 * message is untrusted.
 **/
const char *pb_ntop(int family, const std::string &addr, char *buf, size_t buflen)
{
    if (addr.size() == addrlen(family))
	return inet_ntop(family, addr.data(), buf, buflen);
    unsigned off = snprintf(buf, buflen, "(bad addr/%ld: ", long(addr.size()));
    for (unsigned i = 0; i < addr.size(); i++) {
	if (off > buflen - 5 && i < addr.size() - 2) {
	    off += snprintf(buf+off, buflen-off, "...");
	    break;
	}
	off += snprintf(buf+off, buflen-off, "%02x", addr.data()[i]);
    }
    snprintf(buf+off, buflen-off, ")");
    return buf;
}

/**
 * Convenience variant of pb_ntop() that writes to a static buffer.
 **/
const char *pb_ntop(int family, const std::string &addr)
{
    static char buf[INET6_ADDRSTRLEN+1];
    return pb_ntop(family, addr, buf, sizeof(buf));
}

// Not necessarily the absolute maximum, but a typical maximum.
static inline int addrStrWidth(int family) {
    switch (family) {
	case AF_INET: return 15;
	case AF_INET6: return 32;
	default: return 0;
    }
}

static void printSpoofSchedule(int indent, const SpooferSpoofSchedule &sched,
    int family, bool verboseFlag)
{
    static char buf[SMALLBUF];
    printf("#%*s Spoof schedule (targets: %d)\n", indent, "", sched.item_size());
    if (verboseFlag) {
        for (int i = 0; i < sched.item_size(); i++) {
	    const SpooferSpoofSchedule::Item &item = sched.item(i);
	    time_t ts = item.timestamp();
	    struct tm *tm = gmtime(&ts);
	    strftime(buf, SMALLBUF, "%Y-%m-%d %H:%M:%S", tm);
	    printf("#%*s [%d] %s  ", indent+1, "", i, buf);
	    if (family == AF_INET6)
		printf("\n#%*s   ", indent+1, "");
	    printf("%-*s ", addrStrWidth(family), pb_ntop(family, item.srcip()));
	    printf("-> %-*s\n", addrStrWidth(family), pb_ntop(family, item.dstip()));
	    {
		printf("#%*s   seq: %s", indent+1, "", item.seqno().c_str());
		if (item.has_hmac() && item.hmac().size() > 0) {
		    printf("  HMAC: ");
		    binDump(item.hmac().data(), item.hmac().size(), FALSE);
		}
		printf("\n");
	    }
        }
    }
}

static void printTracefilterSchedule(int indent, const SpooferTracefilterSchedule &sched,
    int family)
{
    printf("#%*s Tracefilter schedule (targets: %d)\n", indent, "", sched.item_size());
    for (int i = 0; i < sched.item_size(); i++) {
	const SpooferTracefilterSchedule::Item &item = sched.item(i);
        printf("#%*s  [%d] dist: %d  ", indent, "", i, item.dist());
	if (family == AF_INET6)
	    printf("\n#%*s    ", indent, "");
	printf("%-*s ", addrStrWidth(family), pb_ntop(family, item.srcip()));
	printf("-> %-*s\n", addrStrWidth(family), pb_ntop(family, item.dstip()));
    }
}

static void printTracerouteSchedule(int indent, const SpooferTracerouteSchedule &sched,
    int family)
{
    printf("#%*s Traceroute schedule (targets: %d)\n", indent, "", sched.item_size());
    for (int i = 0; i < sched.item_size(); i++) {
	const SpooferTracerouteSchedule::Item &item = sched.item(i);
        printf("#%*s  [%d] %s\n", indent, "", i, pb_ntop(family, item.dstip()));
    }
}

void printServerMsg(const SpooferServerMsg &msg, IPv ipv)
{
    int family = ipvtofamily(ipv);
    printf("# ServerMessage (IPv%d):\n", ipv);
    if (msg.has_txtmsg()) {
	printf("#  txtmsg: level %d: \"%s\"\n", msg.txtmsg().level(),
	    msg.txtmsg().body().c_str());
    }
    if (msg.has_hello()) {
	printf("#  hello:\n");
	printf("#   sessionid: %d\n", msg.hello().sessionid());
	printf("#   sessionkey: %s\n", msg.hello().sessionkey().c_str());
	printf("#   clientip: %s\n", pb_ntop(family, msg.hello().clientip()));
    }
    for (int i = 0; i < msg.schedule_size(); i++) {
	printf("#  schedule[%d]:\n", i);
	// schedule holds a oneof, but we avoid oneof API for 2.4 compatibility
	int n_sched = 0;
	if (msg.schedule(i).has_spoof()) {
	    printSpoofSchedule(2, msg.schedule(i).spoof(), family, true);
	    n_sched++;
	}
	if (msg.schedule(i).has_tracefilter()) {
	    printTracefilterSchedule(2, msg.schedule(i).tracefilter(), family);
	    n_sched++;
	}
	if (msg.schedule(i).has_traceroute()) {
	    printTracerouteSchedule(2, msg.schedule(i).traceroute(), family);
	    n_sched++;
	}
	if (n_sched != 1)
	    printf("#   WARNING: sub-schedule count %d != 1\n", n_sched);
    }
    if (msg.has_summary()) {
	printf("#  summary:\n");
	printf("#   AS: %d\n", msg.summary().clientasn());
	printf("#   privaddr: %d\n", msg.summary().privaddr());
	printf("#   routable: %d\n", msg.summary().routable());
    }
}

const char *reportStatusToStr(SpooferReportStatus status)
{
    static char buf[32];
    switch (status) {
	case ERROR: return "ERROR";
	case DONE:  return "DONE";
	case READY: return "READY";
	default: sprintf(buf, "unknown(%d)", status); return buf;
    }
}

const char *spoofStatusToStr(SpooferSpoofReport::Item::Status status)
{
    static char buf[32];
    switch (status) {
	case SpooferSpoofReport::Item::SENDFAIL: return "SENDFAIL";
	case SpooferSpoofReport::Item::OK:       return "OK";
	// case SpooferSpoofReport::Item::NAT:      return "NAT";
	default: sprintf(buf, "unknown(%d)", status); return buf;
    }
}

const char *summaryResultToStr(SpooferResultSummary::Result result)
{
    static char buf[32];
    switch (result) {
	case SpooferResultSummary::UNKNOWN:   return "unknown";
	case SpooferResultSummary::RECEIVED:  return "received";
	case SpooferResultSummary::BLOCKED:   return "blocked";
	case SpooferResultSummary::REWRITTEN: return "rewritten";
	default: sprintf(buf, "unknown(%d)", result); return buf;
    }
}

static void printSpoofReport(int indent, const SpooferSpoofReport &report, int family)
{
    printf("#%*s Spoof report (targets: %d):\n", indent, "", report.item_size());
    printf("#%*s  status: %s\n", indent, "", reportStatusToStr(report.status()));
    for (int i = 0; i < report.item_size(); i++) {
	const SpooferSpoofReport::Item &item = report.item(i);
	printf("#%*s  [%d] ", indent, "", i);
	printf("%-*s ", addrStrWidth(family), pb_ntop(family, item.srcip()));
	printf("-> %-*s", addrStrWidth(family), pb_ntop(family, item.dstip()));
	if (family == AF_INET6)
	    printf("\n#%*s    ", indent, "");
	printf("  seq: %s  status: %s\n", item.seqno().c_str(),
	    spoofStatusToStr(item.status()));
    }
}

static void printTracefilterReport(int indent, const SpooferTracefilterReport &report, int family UNUSED)
{
    printf("#%*s Tracefilter report:\n", indent, "");
    printf("#%*s  status: %s\n", indent, "", reportStatusToStr(report.status()));
}

static void printTracerouteReport(int indent, const SpooferTracerouteReport &report, int family)
{
    printf("#%*s Traceroute report (targets: %d):\n", indent, "", report.item_size());
    printf("#%*s  status: %s\n", indent, "", reportStatusToStr(report.status()));
    for (int i = 0; i < report.item_size(); i++) {
	const SpooferTracerouteReport::Item &item = report.item(i);
	printf("#%*s  [%d] %s\n", indent, "", i, pb_ntop(family, item.dstip()));
	if (item.has_text()) {
	    // preprend the prefix to every line of text
	    const char *p = item.text().c_str();
	    while (*p) {
		int len = strcspn(p, "\n");
		printf("#%*s  %.*s\n", indent+2, "", len, p);
		if (!p[len]) break; // no newline at end
		p += len + 1;
	    }
	}
    }
}

static const char *testtype_to_string(SpooferTestType t)
{
    static char buf[32];
    switch (t) {
	case SPOOF: return "SPOOF";
	case TRACEFILTER: return "TRACEFILTER";
	case TRACEROUTE: return "TRACEROUTE";
	case SPOOFINGRESS: return "SPOOFINGRESS";
	default: sprintf(buf, "unknown(%d)", t); return buf;
    }
}

void printClientMsg(const SpooferClientMsg &msg, IPv ipv)
{
    int family = ipvtofamily(ipv);
    printf("# ClientMessage (IPv%d):\n", ipv);
    printf("#  ready: %s\n", msg.ready() ? "true" : "false");
    if (msg.has_hello()) {
	printf("#  hello:\n");
	printf("#   version: %d\n", msg.hello().version());
	printf("#   os: %s\n", msg.hello().os().c_str());
	printf("#   clientip: %s\n", pb_ntop(family, msg.hello().clientip()));
	printf("#   share_public: %d\n", msg.hello().share_public());
	printf("#   share_remedy: %d\n", msg.hello().share_remedy());
	printf("#   sessionkey: %s\n", msg.hello().sessionkey().c_str());
	printf("#   types: ");
	for (int i = 0; i < msg.hello().type_size(); i++) {
	    printf(" %s", testtype_to_string(msg.hello().type(i)));
	}
	printf("\n");
    }
    for (int i = 0; i < msg.report_size(); i++) {
	printf("#  report[%d]:\n", i);
	// report holds a oneof, but we avoid oneof API for 2.4 compatibility
	int n_report = 0;
	if (msg.report(i).has_spoof()) {
	    printSpoofReport(2, msg.report(i).spoof(), family);
	    n_report++;
	}
	if (msg.report(i).has_tracefilter()) {
	    printTracefilterReport(2, msg.report(i).tracefilter(), family);
	    n_report++;
	}
	if (msg.report(i).has_traceroute()) {
	    printTracerouteReport(2, msg.report(i).traceroute(), family);
	    n_report++;
	}
	if (n_report != 1)
	    printf("#   WARNING: sub-report count %d != 1\n", n_report);
    }
}

