/* 
 * Copyright 2004-2009 Rob Beverly
 * Copyright 2015-2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

/****************************************************************************
   Author:      Rob Beverly <rbeverly at csail.mit.edu>
   Date:        $Date: 2016/05/18 19:34:50 $
   Description: Traceroute wrapper
****************************************************************************/
#ifndef _WIN32

#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include "spoof.h"
#include "prober.h"

static const char cvsid[] USED = "$Id: trace.cc,v 1.14 2016/05/18 19:34:50 kkeys Exp $";

#define MAXWAIT 60  // wait at most this long for traceroute to finish

typedef enum tr_type {UNKNOWN, STANDARD, MISSING, PARALLEL} tr_type;

int n_children = 0;
pid_t pid[MAX_TRACEROUTE_CHILD];
static const char *tracecmd = NULL;
static enum tr_type tracetype = UNKNOWN;

static int checkTraceroute(const char *name, const char *options) {
    FILE *fp;
    char line[BUFSIZE];
    char buf[BUFSIZE];
    int worked = 0;

    /* try it */
    sprintf(buf, "%s %s -n 127.0.0.1 2>&1", name, options);
    fp = popen(buf, "r");
    if (!fp) return 0;
    while (fgets(line, BUFSIZE, fp)) {
	if (strstr(line, "to 127.0.0.1")) {
	    worked = 1;
	    break;
	}
    }
    pclose(fp);
    return worked;
}

/**
 * findtr - Locate traceroute binary on system
 *
 * @param[out]  type    type of traceroute binary
 * @return              path to traceroute binary
 */
static void findTraceroute(enum tr_type &type) {
    static const char *candidate[] = { 	
        "/usr/local/sbin/traceroute",
        "/usr/sbin/traceroute",
        "/sbin/traceroute",
        "/usr/bin/traceroute",
        "/usr/sbin/tracepath",
	NULL,
        };	
    type = MISSING;

    for (int i = 0; candidate[i]; i++) {
	if (checkTraceroute(candidate[i], "")) {
	    tracecmd = candidate[i];
	    printf("traceroute: %s", candidate[i]);
            type = STANDARD;
	    /* check for -N flag common in Linux traceroutes */
	    if (checkTraceroute(candidate[i], "-N1")) {
		printf(" -N1");
		type = PARALLEL;
	    }
	    printf("\n");
	    return;
        }
    }
}

/* Terminate (long running) child */
static void killchild(int child, const char *reason) {
#if DEBUG
    printf("Terminating child %d trace PID %d (%s)\n", child, pid[child], reason);
#endif
    kill(pid[child], SIGKILL);
    int status;
    waitpid(pid[child], &status, WNOHANG);
}

int initTraceroute()
{
    if (tracetype == UNKNOWN)
        findTraceroute(tracetype);
    return tracetype != MISSING;
}

/**
 * runTraceroute - perform a traceroute to a destination 
 *
 * @param[in]   dest        destination
 * @param[out]  result      traceroute output
 *
 * Performs a traceroute to a destination.  Finds the
 * traceroute binary, starts time-limited trace.  Will
 * quit after MAXWAIT seconds
 */
int runTraceroute(Tracker *tracker, struct in_addr *dest, char *result, int num) {
    int bytes_read[MAX_TRACEROUTE_CHILD];
    int progress[MAX_TRACEROUTE_CHILD];
    int pipefd[MAX_TRACEROUTE_CHILD][2];
    struct timeval timeout;
    int start_time = 0;
    int nfds = 0;
    fd_set fds;

    if (!initTraceroute())
        return -1;

    if (num > MAX_TRACEROUTE_CHILD)
        num = MAX_TRACEROUTE_CHILD;

    n_children = 0;
#if DEBUG
    for (int i = 0; i < num; i++)
        printf(">> Running Trace (please wait): %s to %s\n", tracecmd, 
            inet_ntoa(dest[i]));
#endif

    start_time = time(NULL);

    /* open a pipe.  [0] is read end, [1] is write end. */
    for (int i = 0; i < num; i++) {
        if (pipe(pipefd[i]) < 0)
            perror("pipe");
    }

    for (int i = 0; i < num; i++) {
        /* spawn a process to perform the trace */
        switch (pid[i]=fork()) {
            case -1:
                return -1;
            case 0: //child
#if DEBUG
                printf(">> Child ID: %d tracing to: %s\n", i, inet_ntoa(dest[i]));
#endif
		close(pipefd[i][0]); // close unused end
                dup2(pipefd[i][1], STDERR_FILENO);
                dup2(pipefd[i][1], STDOUT_FILENO);
                if (tracetype == PARALLEL)
                    execl(tracecmd, tracecmd, "-n", "-N1", inet_ntoa(dest[i]), (char *) NULL);
                else
                    execl(tracecmd, tracecmd, "-n", inet_ntoa(dest[i]), (char *) NULL);
                exit(0);
            default: //parent 
#if DEBUG
                printf("Parent spawned child %d with pid %d\n", i, pid[i]);
#endif
		close(pipefd[i][1]); // close unused end
                n_children++; 
        }
    } 

    for (int i = 0; i < n_children; i++) {
        bytes_read[i] = 0;
        progress[i] = -1; // start at -1 to skip the first line of output
    }

    /* while traceroutes haven't run too long */
    while (time(NULL) - start_time < MAXWAIT) {
        /* select reset */
        FD_ZERO(&fds);
	nfds = 0;
	for (int i = 0; i < n_children; i++) {
	    if (pipefd[i][0] == -1) continue;
            FD_SET(pipefd[i][0], &fds);
	    if (pipefd[i][0] >= nfds)
		nfds = pipefd[i][0] + 1;
        }
	if (nfds == 0) break; // all pipes are closed
	timeout.tv_sec = start_time + MAXWAIT + 1 - time(NULL);
        timeout.tv_usec = 0;

        /* find a pipe with pending data from which to read */
	int n_ready = select(nfds, &fds, (fd_set *) NULL, (fd_set *) NULL, &timeout);
        if (n_ready < 0) break;
        if (n_ready == 0) continue;

        /* read from pipes which have pending data */
	for (int i = 0; i < n_children; i++) {
            if (pipefd[i][0] == -1) continue; // pipe is closed
	    char *buf = result + i*BIGBUF;
            if ( FD_ISSET(pipefd[i][0], &fds) ) {
                if (bytes_read[i] >= BIGBUF - 1)
                    break;
                int n = read(pipefd[i][0], buf + bytes_read[i], BIGBUF - bytes_read[i] - 1);
		if (n < 0) { // error
                    killchild(i, "read error");
		    close(pipefd[i][0]);
		    pipefd[i][0] = -1;
		} else if (n == 0) { // EOF
		    close(pipefd[i][0]);
		    pipefd[i][0] = -1;
		    int status;
		    waitpid(pid[i], &status, WNOHANG);
                } else {
		    for (int j = 0; j < n; j++) {
			// Show a progress tick for each non-empty line of
			// traceroute output.
			if ((bytes_read[i] > 0 || j > 0) &&
			    buf[bytes_read[i]+j] == '\n' &&
			    buf[bytes_read[i]+j-1] != '\n' &&
			    progress[i] < TRACEROUTE_ITEM_COST - 1)
			{
			    if (++progress[i] > 0) {
				tracker->showProgress();
			    }
			}
		    }
                    bytes_read[i] += n;
		}
            }
        }
    }

    int total = 0;
    for (int i = 0; i < n_children; i++) {
	/* kill any children that didn't exit properly */
	if (pipefd[i][0] != -1) {
	    killchild(i, "unresponsive");
	    close(pipefd[i][0]);
	    pipefd[i][0] = -1;
	}
	/* terminate strings within the big buffer */
	result[i*BIGBUF + bytes_read[i]] = '\0';
	while (progress[i] < TRACEROUTE_ITEM_COST) {
	    if (++progress[i] > 0) {
		tracker->showProgress();
	    }
	}
	/* return the sum of read bytes across all traces */
        total += bytes_read[i];
    }

    return total;
}

#if 0
int main() {
    struct in_addr *dests;
    char *result;
    int numdests = 3;
    int i;

    result = calloc(numdests, BIGBUF);
    dests  = calloc(numdests, sizeof(struct in_addr));

    inet_aton("64.233.161.99", &dests[0]);
    inet_aton("18.7.22.83", &dests[1]);
    inet_aton("128.61.2.1", &dests[2]);

    runtrace(dests, result, numdests);
    for (i=0;i<numdests;i++) {
        printf("Trace %d: %s\n", i, &result[BIGBUF*i]);
    }
    exit(0);
}
#endif

#endif
